#!/usr/bin/env python

 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

from __future__ import with_statement
import util as u
import fcntl
import os
import msg
import socket
import threading
import SocketServer
import struct
import neighbors as neigh
import time
import signal
import sys
import errno
import logging
from random import sample, shuffle, randint
from shapely.wkt import dumps, loads
from shapely.wkb import dumps as b_dumps, loads as b_loads
from shapely.geometry import Point
import Queue
import forwarding as f
import hmac
from hashlib import sha1

# import global settings
#from alibid_test_cfg import *
#from alibid_cfg import *

if len(sys.argv) < 2:
	cfg = __import__('alibid_cfg')
else:
	cfg = __import__(sys.argv[1])


# gloabal variables
neighbors = neigh.Neighbors()
n_lock = threading.Lock()
q_lock = threading.Lock()
q2_lock = threading.Lock()
d_lock = threading.Lock()
qo_lock = threading.Lock()
is_initialized=False
gossip_timer=None
refresh_timer=None
nlist_req_timer=None
server=None
host_name=None
gps_coords=None
query_id=0
queue_map={}
queue_map2={}
query_ids=[]
query_overhead={}


#Handlers
class ThreadedTCPRequestHandler(SocketServer.BaseRequestHandler):
	def handle(self):
		data = ''
		total_len = None
		while True:
			data += self.request.recv(1024)
			if total_len == None: (total_len, ) = struct.unpack('i', data[:struct.calcsize('i')])
			if len(data) > total_len:
				break

		(msg_type,) = struct.unpack('i', data[struct.calcsize('i'):struct.calcsize('i')*2])

		if msg_type == msg.JOIN_MSG:
			logging.debug('join msg received', extra=logging_map)
			self.handle_join(data)
		elif msg_type == msg.QUERY_MSG:
			logging.debug('query msg received', extra=logging_map)
			self.handle_query(data)
		elif msg_type == msg.GOSSIP_MSG:
			logging.debug('gossip msg received', extra=logging_map)
			self.handle_gossip(data)
		elif msg_type == msg.REQUEST_NLIST_MSG:
			logging.debug('request neighbor list msg received', extra=logging_map)
			self.handle_req_nlist(data)
		elif msg_type == msg.REQUEST_QUERY_MSG:
			logging.debug('request query msg received', extra=logging_map)
			self.handle_req_query(data)
		elif msg_type == msg.QUERY_RESPONSE_MSG:
			logging.debug('query response msg received', extra=logging_map)
			self.handle_query_res(data)
		elif msg_type == msg.DUP_CHECK_REQUEST_MSG:
			logging.debug('duplication check request msg received', extra=logging_map)
			self.handle_dup_req(data)
		elif msg_type == msg.QUERY_CHECK_REQUEST_MSG:
			logging.debug('query check request msg received', extra=logging_map)
			self.handle_query_check_req(data)
		elif msg_type == msg.QUERY_OVERHEAD_REQUEST_MSG:
			logging.debug('query overhead request msg received', extra=logging_map)
			self.handle_query_overhead_req(data)
		elif msg_type == msg.PING_MSG:
			logging.debug('ping msg received', extra=logging_map)
			self.handle_ping(data)
		else:
			logging.warning('not supported msg received', extra=logging_map)

	# handle ping (with nonce) message
	def handle_ping(self, data):
		ping_msg = msg.PingMsg()
		(nonce,) = ping_msg.unmarshal(data)

		# send back pong message with nonce in the ping
		pong_msg = msg.PongMsg(nonce)
		self.request.sendall(pong_msg.marshal())

	# check if the query id in the packet has been seen.
	def handle_dup_req(self, data):
		global query_ids

		dq_msg = msg.DupCheckRequestMsg()
		(query_id,) = dq_msg.unmarshal(data)

		is_dup = 0

		with d_lock:
			if query_id in query_ids:
				is_dup=1

		dr_msg = msg.DupCheckReplyMsg(query_id, is_dup)
		#dr_msg = msg.DupCheckReplyMsg(query_id, 0)
		self.update_query_overhead(1, len(dr_msg.marshal()), query_id, 2) #2 for can_handle msgs
		self.request.sendall(dr_msg.marshal())

	# check if the response to the query id in the packet has been received
	def handle_query_check_req(self, data):
		global queue_map2

		dq_msg = msg.QueryCheckRequestMsg()
		(query_id,) = dq_msg.unmarshal(data)

		is_active = 1

		with q2_lock:
			if not queue_map2.get(query_id, False):
				is_active=0

		dr_msg = msg.QueryCheckReplyMsg(query_id, is_active)
		#dr_msg = msg.DupCheckReplyMsg(query_id, 0)
		self.update_query_overhead(1, len(dr_msg.marshal()), query_id, 2) #2 for can_handle msgs
		self.request.sendall(dr_msg.marshal())

	# called when receiving a response for a query
	def handle_query_res(self, data):
		qr_msg = msg.QueryResponseMsg()
		(query_id, succeeded, res_msg, path) = qr_msg.unmarshal(data)

		logging.info('QUERYID %d;Query Response Msg;%d;%s', query_id, int(succeeded), res_msg, extra=logging_map)
	
		with q_lock:
			queue_map[int(query_id)].put((succeeded, res_msg, path))

	# update total number and length of messages sent
	def update_query_overhead(self, num_msg, num_bytes, query_id, tag):
		global query_overhead

		with qo_lock:
			if not query_overhead.has_key(query_id):
				query_overhead[query_id] = [[0,0], [0,0], [0,0]]

			query_overhead[query_id][tag][0]+=num_msg
			query_overhead[query_id][tag][1]+=num_bytes

	# select two next hops (query forking)
	def choose_two_next_nodes(self, safe_nodes, query_id):
		#split nodes into two set
		list_even = []
		list_odd = []
		for i in xrange(len(safe_nodes)):
			if i % 2 == 0:
				list_even.append(safe_nodes[i])
			else:
				list_odd.append(safe_nodes[i])

		#sort them each
		list_even.sort(key=lambda x: x[1])
		list_odd.sort(key=lambda x: x[1])

		best_even_node = None
		best_odd_node = None

		dq_msg = msg.DupCheckRequestMsg(query_id)
		dr_msg = msg.DupCheckReplyMsg()

		sent_msgs=0
		sent_bytes=0

		#asking if they can handle the query
		for n in list_even:
			logging.info('QUERYID %d;Checking if even node %s can handle the query', query_id, n, extra=logging_map)
			sent_msgs+=1
			sent_bytes+=len(dq_msg.marshal())
			response = u.send_msg(n[0], cfg.ALIBI_PORT, dq_msg, True, True)

			if response!=None:
				(query_id, ans) = dr_msg.unmarshal(response)
				if int(ans) == 0:
					logging.info('QUERYID %d;Even node %s can handle the query', query_id, n, extra=logging_map)
					best_even_node = n[0]
					break

		for n in list_odd:
			logging.info('QUERYID %d;Checking if odd node %s can handle the query', query_id, n, extra=logging_map)
			sent_msgs+=1
			sent_bytes+=len(dq_msg.marshal())
			response = u.send_msg(n[0], cfg.ALIBI_PORT, dq_msg, True, True)

			if response!=None:
				(query_id, ans) = dr_msg.unmarshal(response)
				if int(ans) == 0:
					logging.info('QUERYID %d;Odd node %s can handle the query', query_id, n, extra=logging_map)
					best_odd_node = n[0]
					break

		self.update_query_overhead(sent_msgs, sent_bytes, query_id, 2) #2 for can_handle msgs

		return (best_even_node, best_odd_node)

	# called when receiving a query
	def handle_query(self, data):
		global query_ids

		q_msg = msg.QueryMsg()
		q_info = q_msg.unmarshal(data)
		(cur_ip, cur_port) = self.request.getsockname()

		(query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, fr, tr, prev_path, src_host_name, dst_host_name) = q_info


		with d_lock:
			if query_id in query_ids:
				return

			if not cfg.IS_TEST:
				query_ids.append(query_id)


		path = eval(prev_path)+[(host_name, gps_coords[0], gps_coords[1])]

		logging.info('QUERYID %d;SRC %s;DST %s;STRATEGY %d;Hop Counts %d;ttl %d;Query Received', query_id, src_host_name, dst_host_name, strategy, len(path), ttl, extra=logging_map)

		if len(path) > ttl:
			if ttl == cfg.TTL_LIST[-1]:
				if strategy == 1:
					error='0|TTLExpired(@%s)'%(host_name)
					qr_msg = msg.QueryResponseMsg(query_id, 0, error, str(path))
					self.update_query_overhead(1, len(qr_msg.marshal()), query_id, 1) #1 for query_response msgs
					u.send_msg(src_host_name, cfg.ALIBI_PORT, qr_msg)
				return
			else:
				# ask if the search can be kept going
				dq_msg = msg.QueryCheckRequestMsg(query_id)
				dr_msg = msg.QueryCheckReplyMsg()
				response = u.send_msg(src_host_name, cfg.ALIBI_PORT, dq_msg, True, True)

				if response!=None:
					(query_id, ans) = dr_msg.unmarshal(response)
					if int(ans) == 0:
						logging.info('QUERYID %d;SRC %s;DST %s;STRATEGY %d;TTL Expired. Can Keep Going. But Do Not;%d', query_id, src_host_name, dst_host_name, strategy, ttl, extra=logging_map)
						return
				else:
					logging.info('QUERYID %d;SRC %s;DST %s;STRATEGY %d;TTL Expired. Can Keep Going. But Do Not;%d', query_id, src_host_name, dst_host_name, strategy, ttl, extra=logging_map)
					return

				for i in xrange(len(cfg.TTL_LIST)):
					if cfg.TTL_LIST[i] == ttl:
						new_ttl = cfg.TTL_LIST[i+1]
						break

				logging.info('QUERYID %d;SRC %s;DST %s;STRATEGY %d;TTL Expired. But Keep Going.;%d;%d', query_id, src_host_name, dst_host_name, strategy, ttl, new_ttl, extra=logging_map)

				ttl = new_ttl
				


		logging.info('QUERYID %d;SRC %s;DST %s;STRATEGY %d;Processing Query', query_id, src_host_name, dst_host_name, strategy, extra=logging_map)

		# if inside the target region, send back query reply msg to source
		if Point(gps_coords[0], gps_coords[1]).intersects(tr):
			logging.info('QUERYID %d;SRC %s;DST %s;%s within relayzone', query_id, src_host_name, dst_host_name, host_name, extra=logging_map)

			qr_msg = msg.QueryResponseMsg(query_id, 1, host_name, str(path))
			self.update_query_overhead(1, len(qr_msg.marshal()), query_id, 1) #1 for query_response msgs
			u.send_msg(src_host_name, cfg.ALIBI_PORT, qr_msg)
			return

		tar = f.check_nlist_within(query_id, host_name, gps_coords[0], gps_coords[1], neighbors.n_list, neighbors.nn_list, fr, tr)
		if tar != None:
			qr_msg = msg.QueryResponseMsg(query_id, 1, tar, str(path))
			self.update_query_overhead(1, len(qr_msg.marshal()), query_id, 1) #1 for query_response msgs
			u.send_msg(src_host_name, cfg.ALIBI_PORT, qr_msg)
			return

		# get next hop
		prev_hop = self.request.getpeername()[0]

		# find nodes that satisfy safety condition
		safe_nodes = f.get_next_hop(query_id, host_name, src_lon, src_lat, dst_lon, dst_lat, fr, tr, gps_coords[0], gps_coords[1], neighbors.n_list, neighbors.nn_list, prev_hop, path, strategy)

		if len(safe_nodes) == 0:
			error='0|NoValidNextHop(@%s)'%(host_name)
			qr_msg = msg.QueryResponseMsg(query_id, 0, error, str(path))
			self.update_query_overhead(1, len(qr_msg.marshal()), query_id, 1) #1 for query_response msgs
			u.send_msg(src_host_name, cfg.ALIBI_PORT, qr_msg)
			return

		if strategy == 1:
			next_hops = (safe_nodes[0],None)
		elif strategy == 0:
			next_hops = self.choose_two_next_nodes(safe_nodes, query_id)

		if next_hops == (None, None):
			error='0|NoValidNextHop(@%s)'%(host_name)
			qr_msg = msg.QueryResponseMsg(query_id, 0, error, str(path))
			self.update_query_overhead(1, len(qr_msg.marshal()), query_id, 1) #1 for query_response msgs
			u.send_msg(src_host_name, cfg.ALIBI_PORT, qr_msg)
			return

		# forward the query
		for next_hop in next_hops:
			if next_hop == None:
				continue
	
			logging.info('QUERYID %d;SRC %s;DST %s;Trying to forward from %s to %s', query_id, src_host_name, dst_host_name, cur_ip ,next_hop, extra=logging_map)

			s_msg = msg.QueryMsg(query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, fr, tr, str(path), src_host_name, dst_host_name)

			self.update_query_overhead(1, len(s_msg.marshal()), query_id, 0) #0 for query_request msgs
			if not u.send_msg(next_hop, cfg.ALIBI_PORT, s_msg):
				logging.info('QUERYID %d;SRC %s;DST %s;Query forwarding to %s failed', query_id, src_host_name, dst_host_name, next_hop, extra=logging_map)

				with n_lock:
					neighbors.remove_item_from_nlist(next_hop)
					neighbors.update_neighbor_list()

				self.handle_query(data)
				break

	# called when receiving a query from the client module
	def handle_req_query(self, data):
		global query_id, queue_map, query_ids, queue_map2

		while not is_initialized:
			logging.debug('not initialized', extra=logging_map)
			time.sleep(5)

		rq_msg = msg.RequestQueryMsg()
		(m_query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, fr, tr, src_host_name, dst_host_name) = rq_msg.unmarshal(data)


		if m_query_id in query_ids:
			logging.info('QUERYID %d;SRC %s;DST %s;STRATEGY %d;ttl %d;Query Discarded', m_query_id, src_host_name, dst_host_name, strategy, ttl, extra=logging_map)
			return

		if not cfg.IS_TEST:
			with d_lock:
				query_ids.append(m_query_id)

		# create query msg
		#with q_lock:
		#	my_query_id = query_id
		#	query_id+=1

		path = str([(src_host_name, src_lon, src_lat)])
		q_msg = msg.QueryMsg(m_query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, fr, tr, path, src_host_name, dst_host_name)

		logging.info('QUERYID %d;STRATEGY %d;SRC %s;DST %s;Generated at the source', m_query_id, strategy, src_host_name, dst_host_name, extra=logging_map)

		# if inside the target region, send back query reply msg to source
		if Point(gps_coords[0], gps_coords[1]).intersects(tr):
			logging.info('QUERYID %d;SRC %s;DST %s;%s within relayzone', m_query_id, src_host_name, dst_host_name, host_name, extra=logging_map)

			qr_msg = msg.QueryResponseMsg(m_query_id, 1, host_name, '[]')
			self.update_query_overhead(0, 0, m_query_id, 1) #1 for query_response msgs
			self.request.sendall(qr_msg.marshal())
			return

		tar = f.check_nlist_within(m_query_id, host_name, gps_coords[0], gps_coords[1], neighbors.n_list, neighbors.nn_list, fr, tr)
		if tar != None:
			qr_msg = msg.QueryResponseMsg(m_query_id, 1, tar, '[]')
			self.update_query_overhead(0, 0, m_query_id, 1) #1 for query_response msgs
			self.request.sendall(qr_msg.marshal())
			return

		# get next hop
		safe_nodes = f.get_next_hop(m_query_id, host_name, src_lon, src_lat, dst_lon, dst_lat, fr, tr, src_lon, src_lat, neighbors.n_list, neighbors.nn_list, STRATEGY=strategy)

		if len(safe_nodes) == 0:
			error='0|NoValidNextHop(@%s)'%(host_name)
			qr_msg = msg.QueryResponseMsg(m_query_id, 0, error, '[]')
			self.update_query_overhead(0, 0, m_query_id, 1) #1 for query_response msgs
			self.request.sendall(qr_msg.marshal())
			return

		if strategy == 1:
			next_hops = (safe_nodes[0], None)
		elif strategy == 0:
			next_hops = self.choose_two_next_nodes(safe_nodes, m_query_id)

		if next_hops == (None, None):
			error='0|NoValidNextHop(@%s)'%(host_name)
			qr_msg = msg.QueryResponseMsg(m_query_id, 0, error, '[]')
			self.update_query_overhead(0, 0, m_query_id, 1) #1 for query_response msgs
			self.request.sendall(qr_msg.marshal())
			return

		q = Queue.Queue()
		#queue_map[my_query_id] = q
		with q_lock:
			queue_map[m_query_id] = q

		with q2_lock:
			queue_map2[m_query_id] = True

		for next_hop in next_hops:
			if next_hop == None:
				continue
			logging.info('QUERYID %d;SRC %s;DST %s;Trying to forward from the source to %s', m_query_id, src_host_name, dst_host_name, next_hop, extra=logging_map)

			self.update_query_overhead(1, len(q_msg.marshal()), m_query_id, 0) #0 for query_request msgs
			# u.send_msg
			u.send_msg(next_hop, cfg.ALIBI_PORT, q_msg)

		# wait for response(s) to the query.
		while True:
			try:
				res = q.get(True, cfg.QUERY_TIMEOUT)
			except:
				logging.debug('QUERYID %d;SRC %s;DST %s;Query timed out', m_query_id, src_host_name, dst_host_name, extra=logging_map)
				res = None
				error='0|QueryTimedOut'
				qr_msg = msg.QueryResponseMsg(m_query_id, 0, error, '[]')
				break

			if res != None:
				(succeeded, res_msg, path) = res
				qr_msg = msg.QueryResponseMsg(m_query_id, succeeded, res_msg, path)
				if succeeded == 1:
					logging.info('QUERYID %d;SRC %s;DST %s;Got a query response (Relay - %s)|%s', m_query_id, host_name, dst_host_name, res_msg, path, extra=logging_map)
					#res_msg = '1|'+res_msg
					break
				else:
					logging.info('QUERYID %d;SRC %s;DST %s;Query Failed;%s;%s', m_query_id, host_name, dst_host_name, res_msg, path, extra=logging_map)
					# if strategy == 1 (that is, query forking is not used), then at most one peer sends a response. Therefore, stop waiting for response if a response has been received. Otherwise, multiple peers may send responses thus keep waiting for response.
					if strategy == 1:
						break

		with q2_lock:
			queue_map2[m_query_id] = False
		with q_lock:
			q.queue.clear()

		self.update_query_overhead(1, len(qr_msg.marshal()), m_query_id, 1) #1 for query_response msgs


		client = self.request.getpeername()[0]
		logging.info('QUERYID %d;Send a query response to %s', m_query_id, client, extra=logging_map)
		# send response back to the client
		self.request.sendall(qr_msg.marshal())

	# called when a query overhead request message is received.
	def handle_query_overhead_req(self, data):
		global query_overhead
	
		logging.info('handling query_overhead_req msg', extra=logging_map)

		# send back query overhead
		with qo_lock:
			qo_res_msg = msg.QueryOverheadResponseMsg(str(query_overhead))

		self.request.sendall(qo_res_msg.marshal())

	# called when another peer requests neighbor list
	def handle_req_nlist(self, data):
		(ip, port) = self.request.getpeername()
		(cur_ip, cur_port) = self.request.getsockname()
		logging.info('handling req_nlist msg from %s', ip, extra=logging_map)
		rnlist_msg = msg.RequestNlistMsg()
		(lon, lat) = rnlist_msg.unmarshal(data)
		# repurpose lon and lat to represent diverse demands
		# if lon <= -1100 => request neighbors with latency information
		# if lon > -500 => request neighbors and known-active peers without latency information
		# if -1100 < lon <= -500 and lat == 0 => request neighbors without latency information
		# if -1100 < lon <= -500 and lat != 0 => request known-active peers without latency information

		if lon > -500:
			# add the sender to known-active-peers
			node_list = neighbors.add_latency([(ip, lon, lat)])
			with n_lock:
				neighbors.add_to_candidate_list(node_list)
				neighbors.update_neighbor_list()

				# send back the gossip msg (containing neighbors and known-active peers)
				send_buf = neighbors.marshal(neighbors.c_list+neighbors.n_list+[(cur_ip, gps_coords[0], gps_coords[1], 0.0)])
		elif lon > - 1100:
			with n_lock:
				if lat == 0:
					# send back the gossip msg (containing neighbors)
					send_buf = neighbors.marshal(neighbors.n_list)
				else:
					# send back the gossip msg (containing known-active peers)
					send_buf = neighbors.marshal(neighbors.c_list)
		else:
			with n_lock:
				# send back the gossip msg (containing neighbors with latency information)
				send_buf = neighbors.marshal_wl(neighbors.n_list)
	
		gossip_msg = msg.GossipMsg(send_buf)
		self.request.sendall(gossip_msg.marshal())

	# called when first joining the system
	def handle_join(self, data):
		global is_initialized, gossip_timer, refresh_timer, nlist_req_timer

		if not is_initialized:
			cur_lon, cur_lat = map(float, gps_coords)
			nlist_msg = msg.RequestNlistMsg(cur_lon, cur_lat)

			while True:
				# keep trying to connect to bootstrapping nodes
				for b_node in sample(cfg.bootstrapping_nodes, len(cfg.bootstrapping_nodes)):
					if host_name == b_node:
						continue

					# send reguest neighbor list msg to the bootstrap node
					logging.info('Trying to connect a bootstrapping node;%s', b_node, extra=logging_map)
					response = u.send_msg(b_node, cfg.ALIBI_PORT, nlist_msg, True, True)

					if response!=None:
						cur_ip = socket.gethostbyname(socket.gethostname())
						# response is a gossip msg
						self.handle_gossip(response, b_node, cur_ip)

						if len(neighbors.n_list) < 10:
							continue


						logging.info('connected to the bootstrapping node %s', b_node, extra=logging_map)

						# schedule gossipping
						gossip_timer = threading.Timer(cfg.GOSSIP_INTERVAL+randint(-1,1), self.send_gossip_msg)
						gossip_timer.start()

						refresh_timer = threading.Timer(cfg.REFRESH_INTERVAL, self.refresh_neighbors)
						refresh_timer.start()

						nlist_req_timer = threading.Timer(cfg.NLIST_REQ_INTERVAL+randint(-1,1), self.refresh_two_hop_info)
						nlist_req_timer.start()

						is_initialized = True
						break

				if is_initialized:
					break

				time.sleep(5)

	# called when nlist_req_timer is expired
	def refresh_two_hop_info(self):
		global nlist_req_timer

		logging.info('Refresh two-hop info', extra=logging_map)

		# update two-hop neighbors information
		self.update_two_hop_info(neighbors.n_list)

		nlist_req_timer = threading.Timer(cfg.NLIST_REQ_INTERVAL+randint(-1,1), self.refresh_two_hop_info)
		nlist_req_timer.start()


	# called when refresh_timer is expired
	def refresh_neighbors(self):
		global refresh_timer

		logging.info('Refresh neighbor lists', extra=logging_map)

		with n_lock:
			c_c_list = neighbors.c_list[:]
			c_c_list += neighbors.n_list[:]

		# measure RTTs
		c_c_list = neighbors.add_latency(c_c_list)

	
		with n_lock:
			neighbors.process_rtt_result(c_c_list)
			# update the neighbor list for diversity
			new_neighbors = neighbors.update_neighbor_list()

		#logging.info('ALLNEIGHBORS;%d;%d;%s;%s;%f;%f', len(neighbors.n_list), len(neighbors.c_list), neighbors.n_list, neighbors.c_list, neighbors.check_rtt_dispersion(), neighbors.check_gps_dispersion(), extra=logging_map)

		# update two-hop neighbors information of new neighbors
		self.update_two_hop_info(new_neighbors)

		refresh_timer = threading.Timer(cfg.REFRESH_INTERVAL, self.refresh_neighbors)
		refresh_timer.start()

	# called when gossip_time is expired
	def send_gossip_msg(self):
		global gossip_timer

		with n_lock:
			# choose one random neighbor
			if len(neighbors.n_list) > 0:
				(n_ip, lon, lat, latency) = sample(neighbors.n_list, 1)[0]
				gossip_num = min(neighbors.gossip_num, len(neighbors.c_list+neighbors.n_list))
				send_buf = neighbors.marshal(sample(neighbors.c_list+neighbors.n_list, gossip_num))
			else:
				send_buf = None

		if send_buf != None:
			logging.info('Send gossip msg to %s', n_ip, extra=logging_map)

			# send a gossip message
			gossip_msg = msg.GossipMsg(send_buf)
			#print neighbors.unmarshal(send_buf)
			if not u.send_msg(n_ip, cfg.ALIBI_PORT, gossip_msg):
				logging.info('Gossip msg send failed. remove %s from neighbor list.', n_ip, extra=logging_map)

				with n_lock:
					neighbors.remove_item_from_nlist(n_ip)
					neighbors.update_neighbor_list()


		gossip_timer = threading.Timer(cfg.GOSSIP_INTERVAL+randint(-1,1), self.send_gossip_msg)
		gossip_timer.start()


	# ask neighbors for their neighbor tables
	def update_two_hop_info(self, t_list):
		gossip_msg = msg.GossipMsg()
		nlist_msg = msg.RequestNlistMsg(-2000, 0) # if lon <= -1100 => request a list of neighbors with latency information

		for n in t_list:
			response = u.send_msg(n[0], cfg.ALIBI_PORT, nlist_msg, True, True)
			if response != None:
				(tmp_buf,) = gossip_msg.unmarshal(response)
				with n_lock:
					neighbors.nn_list[n[0]]=neighbors.unmarshal_wl(tmp_buf)
				
					#logging.info('Two-hop neighbor table obtained;%s;%d;%s', n[0], len(neighbors.nn_list[n[0]]), neighbors.nn_list[n[0]], extra=logging_map)


	# called when a gossip message is received
	def handle_gossip(self, data, peer=None, cur_ip=None):
		if peer==None: peer=self.request.getpeername()[0]
		if cur_ip==None: (cur_ip, cur_port) = self.request.getsockname()

		gossip_msg = msg.GossipMsg()
		(tmp_buf,) = gossip_msg.unmarshal(data)
		a_list = neighbors.unmarshal(tmp_buf)
		logging.info('Received gossip msg from %s;%s', peer, a_list, extra=logging_map)

		if not cfg.IS_TEST:
			a_list = neighbors.remove_item_from_list(cur_ip ,a_list)
		a_list = neighbors.add_latency(a_list)

		logging.info('Remove itself from gossip msg %s;%s', cur_ip, a_list, extra=logging_map)
		with n_lock:
			# update the list of know-active peers
			neighbors.add_to_candidate_list(a_list)
			# update the list of neighbors
			new_neighbors = neighbors.update_neighbor_list()

		# update two-hop neighbors information of new neighbors
		self.update_two_hop_info(new_neighbors)


# SIGINT handler
def sigint_handler(signal, fram):
	gossip_timer.cancel()
	refresh_timer.cancel()
	nlist_req_timer.cancel()
	server.shutdown()
	sys.exit(0)
signal.signal(signal.SIGINT, sigint_handler)



# create server
class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
	def server_bind(self):
		self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.socket.bind(self.server_address)



# set host name and gps coordinates
host_name=socket.gethostbyaddr(socket.gethostname())[0]
gps_coords=u.get_gps_coords(host_name,cfg.GPS_FILE)

if gps_coords == None:
	logging.error('Cannot obtain gps coords - %s', host_name)
	#gps_coords = (0, 0)
	sys.exit(1)

neighbors.set_gps_coords(gps_coords)

# set logging parameters
logging_map={'hostname':host_name}

logging.Formatter.converter = time.gmtime
logging.basicConfig(format='%(asctime)s|%(levelname)s|%(hostname)s|%(message)s', level=cfg.LOGGING_LEVEL)
#logging.basicConfig(format='%(levelname)s|%(message)s', level=logging.DEBUG)


HOST, PORT = 'localhost', cfg.ALIBI_PORT
server = ThreadedTCPServer(('', PORT), ThreadedTCPRequestHandler)



# run server
server_thread = threading.Thread(target=server.serve_forever)
# Exit the server thread when the main thread terminates
server_thread.daemon = True
server_thread.start()


# create the join message
join_msg = msg.JoinMsg()

u.send_msg(HOST, PORT, join_msg)

#request_query_msg = msg.RequestQueryMsg(5, 1.2, 0.5, 'pl1snu.koren.kr', 'BRA')


#res_msg = u.send_msg(HOST, PORT, request_query_msg, True, True)
#print res_msg[struct.calcsize('i'):]

if cfg.IS_TEST:
	raw_input()
else:
	server_thread.join()

#print query_overhead

gossip_timer.cancel()
refresh_timer.cancel()
nlist_req_timer.cancel()
server.shutdown()
sys.exit(0)

