#!/usr/bin/env python

 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

import util as u
import msg
import sys
import time
import numpy as np
from datetime import datetime
from shapely.wkb import dumps, loads

from alibid_cfg import *

import os

sys.path.append(os.path.abspath('../target_region_computation'))

import regions as r

# TTL_LIST is a variable in alibid_cfg.py
ttl=TTL_LIST[0]

# Strategy for query processing. See forward.py for details.
strategy = 0

def create_and_send_query_msg(strategy, src, dst, src_lon, src_lat, dst_lon, dst_lat, tr, fr,query_id):
	if tr == None:
		error='TargetRegionNotFound'
		print error
		return

	rq_msg = msg.RequestQueryMsg(query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, fr, tr, src, dst)

	response = u.send_msg(src, ALIBI_PORT, rq_msg, reply=True, timeout=7)

	if response == None:
		print >> sys.stdout, 'Cannot send to nor receive from %s,%d'%(src,query_id)
	else:
		qr_msg = msg.QueryResponseMsg()
		qr_info = qr_msg.unmarshal(response)
		if qr_info[2][:2] == '0|':
			print >> sys.stdout, 'Failure,%d,%s'%(qr_info[0], qr_info[2])
		else:
			print >> sys.stdout, 'Success,%d,%s'%(qr_info[0], qr_info[2])


query_id = 0

# Read queries from STDIN. See README for the format of queries.
for line in sys.stdin:
	(src, src_lon, src_lat, dst, dst_lon, dst_lat, forbidden_region, delta) = line.split(',')
	ineq_factor = float(delta)+1.0

	fr = None
	with open('./data/forbidden_region_%s.bin.dat'%(forbidden_region), 'rb') as f:
		fr = loads(f.read())

	print >> sys.stderr, 'Generating target region ...'
	tr = r.get_target_region(fr, (src_lon,src_lat), (dst_lon,dst_lat), float(ineq_factor))

	if tr == None:
		continue

	print >> sys.stdout, '# Query ID - %d,%s'%(query_id,line.strip())

	create_and_send_query_msg(strategy, src, dst, float(src_lon), float(src_lat), float(dst_lon), float(dst_lat), tr, fr, query_id)

	query_id+=1
	
	time.sleep(1)


