#!/usr/bin/env python

 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

import string
import simulate
from node import Node
import util
import sys
from shapely.wkb import dumps as b_dumps, loads as b_loads
import os

sys.path.append(os.path.abspath('../target_region_computation'))

import regions as r

forbidden_region = 'CHN'
nnodes = '10k'

coordinate_file = "./data/%s_gps_coords.dat"%(nnodes)
neighbor_file = "./data/%s_ntable.dat"%(nnodes)

nodes = util.loadNodeInfo(coordinate_file, neighbor_file)

dists_to_fr = {}

precomputed_target_regions = r.load_precomputed_trs(os.path.abspath('./data/target_regions_%s.dat'%(forbidden_region)))

with open('./data/forbidden_region_%s.bin.dat'%(forbidden_region), 'rb') as f:
    fr = b_loads(f.read())

dists_to_fr = {}

with open('./data/%s_dists_to_fr_%s.dat'%(nnodes, forbidden_region),'r') as f:
	for line in f:
		(lon, lat, dist) = line.split()
		dists_to_fr[(lon,lat)] = float(dist)*simulate.MILES_PER_METER

strategy=0

ttl_list=[2,4,7]
dc_len=28
qr_len=200
fr_len=len((b_dumps(fr)))

ctr_total=0       # Counter for total cases
ctr_no_path=0  # Counter for 'No valid path' case
hop_counts=[]
paths=[]
msg_num=[[] for _ in xrange(2)]
msg_len=[[] for _ in xrange(2)]
node_num=[[] for _ in xrange(2)]
norm_msg_num=[[] for _ in xrange(2)]
norm_msg_len=[[] for _ in xrange(2)]

line_ctr=0

with open("./data/triples_%s.dat"%(forbidden_region)) as f:
    for line in f:
        (avoid, src_id, src_lon, src_lat, dst_id, dst_lon, dst_lat, ineq_factor) = line.split()

        if line_ctr%200 == 0:
            print >> sys.stderr, avoid, line_ctr
        line_ctr+=1

        #if line_ctr == 5:
        #    break

        ctr_total+=1

        tr = r.get_target_region(fr, (src_lon,src_lat), (dst_lon,dst_lat), float(ineq_factor),precomputed_target_regions)

        ret = simulate.simulate(src_id, dst_id, tr, fr, nodes, strategy, dists_to_fr, None, {}, ttl_list)

        res_idx = 0
        if ret[0]: res_idx = 1

        if strategy == 0:
            msg_num[res_idx].append(2*ret[2]+ret[3]+ret[5])
        elif strategy == 1:
            msg_num[res_idx].append(2*ret[2]+ret[3])
        node_num[res_idx].append(ret[3]+1)

        if strategy == 0:
            msg_len[res_idx].append(2*ret[2]*dc_len+ret[6]+ret[7])
        elif strategy == 1:
            msg_len[res_idx].append(ret[6]+ret[7])
        else:
            print 'fatal error - invalid strategy number:', strategy

        norm_msg_num[res_idx].append(1.*msg_num[res_idx][-1]/node_num[res_idx][-1])
        norm_msg_len[res_idx].append(1.*msg_len[res_idx][-1]/node_num[res_idx][-1])

        if ret[0] == True:
            hop_counts.append(len(ret[1])-1)
            paths.append(ret[1])
        else:
            ctr_no_path+=1
            #sys.stdout.write(line)


'''
print 'paths -', paths
print 'hop count list -', hop_counts
print 'msg num failure list -', msg_num[0]
print 'msg num success list -', msg_num[1]
print 'msg len failure list -', msg_len[0]
print 'msg len success list -', msg_len[1]
print 'node num failure list -', node_num[0]
print 'node num success list -', node_num[1]
'''

(hop_counts_mean, hop_counts_std) = util.compute_mean_and_std(hop_counts)
(msg_num_mean, msg_num_std) = util.compute_mean_and_std(msg_num[0]+msg_num[1])
(msg_len_mean, msg_len_std) = util.compute_mean_and_std(msg_len[0]+msg_len[1])
(node_num_mean, node_num_std) = util.compute_mean_and_std(node_num[0]+node_num[1])

print 'Number of triples in total -', ctr_total

print 'Number of triples failed to find a path -', ctr_no_path
if ctr_total > 0:
    print 'Success rate -', float(ctr_total-ctr_no_path)/ctr_total*100
print 'Average number of hops per query -', hop_counts_mean, 'standard deviation -', hop_counts_std
print 'Average number of transmitted messages per query -', msg_num_mean, 'standard deviation -', msg_num_std
print 'Average total length of transmitted messages per query -', msg_len_mean, 'standard deviation -', msg_len_std 
print 'Average number of contacted nodes per query -', node_num_mean, 'standard deviation -', node_num_std




#print 'norm. avg msg num -', np.mean(norm_msg_num[0]+norm_msg_num[1]), '(standard deviation - )', np.std(norm_msg_num[0]+norm_msg_num[1])
#print 'norm. avg msg len -', np.mean(norm_msg_len[0]+norm_msg_len[1]), '(standard deviation - )', np.std(norm_msg_len[0]+norm_msg_len[1])

