
 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

from shapely.ops import cascaded_union
from shapely.geometry import LineString
from shapely.geometry import Point
from shapely.geometry import MultiPoint
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
from node import Node
import sys
from shapely.wkb import dumps as b_dumps, loads as b_loads
import os

sys.path.append(os.path.abspath('../target_region_computation'))

import regions as r

MILES_PER_METER = 0.000621371        #miles/meter

# Get the physical distance from a point to a multipolygon
def getDistanceToPoly(lon, lat, mpoly):
    return r.dist_point_to_mpoly(mpoly, lon, lat)*MILES_PER_METER #miles

def getLatencyEstimate(distance):
    # TODO Annotate which is the 2/3*c => done
    #SLOPE = 0.0269                     #ms/mile -- the Htrae slope
    #SLOPE = 0.01345                    #ms/mile -- 1/2 of the Htrae slope
    SLOPE = 0.0161                      #ms/mile -- 3/c [Equation 4]

    #INTERCEPT = 63.32                  #ms -- the Htrae intercept
    INTERCEPT = 0.0

    return distance*SLOPE + INTERCEPT   #ms


def simulate(src_id, dst_id, targetRegion, forbiddenRegion, nodes, STRATEGY=0, dists_to_fr=None, ttl=None, dists_to_tr={}, ttl_list=[2,4,7], e_neighbors=None, debug=False):
    src_lon,src_lat = map((lambda array: array[0]), nodes[src_id].coordinate.coords.xy)
    dst_lon,dst_lat = map((lambda array: array[0]), nodes[dst_id].coordinate.coords.xy)

    to_process = list()
    processed = list()

    path = []
    path.append(src_id)
    
    safe_neighbors_per_node = {}
    safe_neighbors_per_node[src_id]=[];

    ttl_idx_per_node = {}

    path_per_node = {}  # routing path per node
    path_per_node[src_id] = [src_id];


    q_msg_ctr = 0	# number of transmitted query messages (by all peers)
    q_msg_len = 0	# total lengh of trasmitted query messages (by all peers)
    qr_msg_ctr = 0	# number of trasmitted query response messages (by all peers)
    qr_msg_len = 0	# total length of transmitted query response meesages (by all peers)
    dc_msg_ctr = 0	# number of transmitted DupCheckRequest and Reply messages (by all peers)
    ret_path = None # a path to return

    q_msg_d_len = 52+len((b_dumps(targetRegion)))+len((b_dumps(forbiddenRegion)))+len(src_id)+len(dst_id) # length of a single query message

    to_process.append(src_id)

    found_someone_in_target = False

    while len(to_process) > 0 and not found_someone_in_target:
        curr_node_id = to_process.pop(0)

        curr_node_lon,curr_node_lat = map((lambda array: array[0]), nodes[curr_node_id].coordinate.coords.xy)

        if dists_to_fr != None:
            dist_to_forbidden = dists_to_fr[nodes[curr_node_id].coordinate_str]
        else:
            dist_to_forbidden = getDistanceToPoly(curr_node_lon, curr_node_lat, forbiddenRegion)
        if dists_to_tr.has_key(curr_node_id):
            dist_to_target = dists_to_tr[curr_node_id]
        else:
            dist_to_target = getDistanceToPoly(curr_node_lon, curr_node_lat, targetRegion)
            dists_to_tr[curr_node_id] = dist_to_target

        min_rtt_to_forbidden = getLatencyEstimate(dist_to_forbidden)

        print >> sys.stderr, '[%s;%s;%d] curr_id - %s, curr_lon - %s, curr_lat - %s, dist_to_forbidden - %f, dist_to_target - %f, min_rtt_to_forbidden - %f'%(src_id, dst_id, STRATEGY, curr_node_id, curr_node_lon, curr_node_lat, dist_to_forbidden, dist_to_target, min_rtt_to_forbidden)

        processed.append(curr_node_id)

        #############################################################################
        # STRATEGY 0 is the strategy described in the paper [Section 4.4].
        if STRATEGY == 0:
            min_node_id = [None]*2
            min_distance = [None]*2
            safe_neighbors=[[] for _ in xrange(2)]
            is_found=False
            idx=0
          
            for neighbor_id in nodes[curr_node_id].neighbors.keys():
                if neighbor_id not in nodes.keys():
                    continue

                if nodes[curr_node_id].neighbors[neighbor_id] >= min_rtt_to_forbidden:
                    continue

                if nodes[neighbor_id].coordinate.intersects(targetRegion):
                    if curr_node_id != src_id:
                        qr_msg_ctr+=1
                        qr_msg_len+=(20+len(neighbor_id)+len(str(path_per_node[curr_node_id])))



                    print >> sys.stderr, '[%s;%s;%d] curr_id - %s, neighbor %s in target region, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id, str(path_per_node[curr_node_id]+[neighbor_id]))


                    if ret_path == None:
                        print >> sys.stderr, '[%s;%s;%d] curr_id - %s, neighbor %s in target region, the shortest one, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id, str(path_per_node[curr_node_id]+[neighbor_id]))
                        ret_path = path_per_node[curr_node_id]+[neighbor_id]
                    else:
                        print >> sys.stderr, '[%s;%s;%d] curr_id - %s, neighbor %s in target region, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id, str(path_per_node[curr_node_id]+[neighbor_id]))


                    is_found=True
                    min_node_id[0] = None
                    min_node_id[1] = None
                    break

                neighbor_lon,neighbor_lat = map((lambda array: array[0]), nodes[neighbor_id].coordinate.coords.xy)

                l_min_node_id = None
                l_min_distance = None
  
                if dists_to_tr.has_key(neighbor_id):
                    l_dist_to_target = dists_to_tr[neighbor_id]
                else:
                    l_dist_to_target = getDistanceToPoly(neighbor_lon, neighbor_lat, targetRegion)
                    dists_to_tr[neighbor_id] = l_dist_to_target

                if dists_to_fr != None:
                    l_min_distance = l_dist_to_target - dists_to_fr[nodes[neighbor_id].coordinate_str]
                    l_min_rtt_to_forbidden = getLatencyEstimate(dists_to_fr[nodes[neighbor_id].coordinate_str])
                else:
                    l_min_distance = l_dist_to_target - getDistanceToPoly(neighbor_lon,neighbor_lat, forbiddenRegion)
                    l_min_rtt_to_forbidden = getLatencyEstimate(getDistanceToPoly(neighbor_lon,neighbor_lat, forbiddenRegion))

                print >> sys.stderr, '[%s;%s;%d] curr_id - %s, curr_lon - %s, curr_lat - %s, dist_to_forbidden - %f, dist_to_target - %f, neighbor_id - %s, neighbor_dist_to_target - %f, neighbor_dist_to_forbidden - %f, metric - %f'%(src_id, dst_id, STRATEGY, curr_node_id, curr_node_lon, curr_node_lat, dist_to_forbidden, dist_to_target, neighbor_id, l_dist_to_target, l_dist_to_target-l_min_distance, l_min_distance)


                break_flag = False

                #for neighbor_id2 in e_neighbors[neighbor_id].keys():
                for neighbor_id2 in nodes[neighbor_id].neighbors.keys():
                    if neighbor_id2 not in nodes.keys():
                        continue

                    #if e_neighbors[neighbor_id][neighbor_id2] < l_min_rtt_to_forbidden:
                    if nodes[neighbor_id].neighbors[neighbor_id2] < l_min_rtt_to_forbidden:
                        if nodes[neighbor_id2].coordinate.intersects(targetRegion):

                            if curr_node_id != src_id:
                                qr_msg_ctr+=1
                                qr_msg_len+=(20+len(neighbor_id2)+len(str(path_per_node[curr_node_id])))


                            if ret_path == None:
                                print >> sys.stderr, '[%s;%s;%d] curr_id - %s, two-hop neighbor %s in target region, shortest one, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id2, str(path_per_node[curr_node_id]+[neighbor_id2]))
                                ret_path = path_per_node[curr_node_id]+[neighbor_id2]
                            else:
                                print >> sys.stderr, '[%s;%s;%d] curr_id - %s, two-hop neighbor %s in target region, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id2, str(path_per_node[curr_node_id]+[neighbor_id2]))

                            #query response msg

                            min_node_id[0] = None
                            min_node_id[1] = None
                            is_found = True
                            break_flag = True
                            break

                        neighbor_lon2,neighbor_lat2 = map((lambda array: array[0]), nodes[neighbor_id2].coordinate.coords.xy)

                        if dists_to_tr.has_key(neighbor_id2):
                            l2_dist_to_target = dists_to_tr[neighbor_id2]
                        else:
                            l2_dist_to_target = getDistanceToPoly(neighbor_lon2, neighbor_lat2, targetRegion)
                            dists_to_tr[neighbor_id2] = l2_dist_to_target

                        if dists_to_fr != None:
                            distance = l2_dist_to_target - dists_to_fr[nodes[neighbor_id2].coordinate_str]
                        else:
                            distance = l2_dist_to_target - getDistanceToPoly(neighbor_lon2,neighbor_lat2, forbiddenRegion)

                        if l_min_distance == None or distance < l_min_distance:
                            l_min_distance = distance
                            l_min_node_id = neighbor_id2


                        print >> sys.stderr, '[%s;%s;%d] curr_id - %s, curr_lon - %s, curr_lat - %s, dist_to_forbidden - %f, dist_to_target - %f, two-hop neighbor_id - %s, two-hop neighbor_dist_to_target - %f, two-hop neighbor_dist_to_forbidden - %f, metric - %f'%(src_id, dst_id, STRATEGY, curr_node_id, curr_node_lon, curr_node_lat, dist_to_forbidden, dist_to_target, neighbor_id2, l2_dist_to_target, l2_dist_to_target-distance, distance)

                if break_flag:
                    break

                if l_min_node_id != None:
                    safe_neighbors[idx].append((neighbor_id, l_min_distance))
                    idx = 1-idx

            if not is_found:
                for i in xrange(2):
                    safe_neighbors[i].sort(key=lambda x: x[1])
                    for n in safe_neighbors[i]:
                        (neighbor_id, distance) = n

                        dc_msg_ctr+=1

                        if neighbor_id in to_process:
                            continue

                        if neighbor_id in processed:
                            continue

                        min_node_id[i] = neighbor_id
                        min_distance[i] = distance
                        break

            while True:
                ttl = ttl_list[ttl_idx_per_node.get(curr_node_id, 0)]
                if ttl == None or len(path_per_node[curr_node_id]) <= ttl:
                    if (min_node_id[0] != None):
                        q_msg_len+=(q_msg_d_len+len(str(path_per_node[curr_node_id])))
                        path_per_node[min_node_id[0]] = path_per_node[curr_node_id]+[min_node_id[0]]
                        q_msg_ctr+=1
                        to_process.append(min_node_id[0])

                    if (min_node_id[1] != None):
                        q_msg_ctr+=1
                        q_msg_len+=(q_msg_d_len+len(str(path_per_node[curr_node_id])))
                        path_per_node[min_node_id[1]] = path_per_node[curr_node_id]+[min_node_id[1]]
                        to_process.append(min_node_id[1])

                    if min_node_id[0] == None and min_node_id[1] == None:
                        qr_msg_ctr+=1
                        qr_msg_len+=(39+len(curr_node_id)+len(str(path_per_node[curr_node_id])))
                        if is_found:
                            print >> sys.stderr, '[%s;%s;%d] curr_id - %s, found a relay node, %s'%(src_id, dst_id, STRATEGY, curr_node_id, str(ret_path[-1]))
                        else:
                            print >> sys.stderr, '[%s;%s;%d] curr_id - %s, no valid next hop, %s'%(src_id, dst_id, STRATEGY, curr_node_id, str(path_per_node[curr_node_id]))

                    break
                elif ttl == ttl_list[-1]:
                    break
                elif ret_path != None:
                    break
                else:
                    qr_msg_ctr+=2
                    qr_msg_len+=(39+len(curr_node_id)+len(str(path_per_node[curr_node_id])))
                    qr_msg_len+=(39+len(curr_node_id)+len(str(path_per_node[curr_node_id])))

                    ttl_idx_per_node[curr_node_id] = ttl_idx_per_node.get(curr_node_id,0)+1


        #############################################################################
        # STRATEGY 1 is a strategy that does not use "query forking" [Section 4.5].
        elif STRATEGY == 1:
            min_node_id = None
            min_distance = dist_to_target - dist_to_forbidden
            
            for neighbor_id in nodes[curr_node_id].neighbors.keys():
                if found_someone_in_target:
                    break
                
                if neighbor_id in processed:
                    continue

                if neighbor_id not in nodes.keys():
                    continue

                if nodes[curr_node_id].neighbors[neighbor_id] < min_rtt_to_forbidden:
                    if nodes[neighbor_id].coordinate.intersects(targetRegion):
                        if curr_node_id != src_id:
                            qr_msg_len+=(20+len(neighbor_id)+len(str(path)))

                        print >> sys.stderr, '[%s;%s;%d] curr_id - %s, neighbor %s in target region, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id, str(path+[neighbor_id]))

                        return True, path+[neighbor_id], 0, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len
                        found_someone_in_target = True
                        processed.append(neighbor_id)
                        break

                    neighbor_lon,neighbor_lat = map((lambda array: array[0]), nodes[neighbor_id].coordinate.coords.xy)

                    l_min_node_id = None


                    if dists_to_tr.has_key(neighbor_id):
                        l_dist_to_target = dists_to_tr[neighbor_id]
                    else:
                        l_dist_to_target = getDistanceToPoly(neighbor_lon, neighbor_lat, targetRegion)
                        dists_to_tr[neighbor_id] = l_dist_to_target


                    if dists_to_fr != None:
                        l_min_distance = l_dist_to_target - dists_to_fr[nodes[neighbor_id].coordinate_str]
                        l_min_rtt_to_forbidden = getLatencyEstimate(dists_to_fr[nodes[neighbor_id].coordinate_str])
                    else:
                        l_min_distance = l_dist_to_target - getDistanceToPoly(neighbor_lon,neighbor_lat, forbiddenRegion)
                        l_min_rtt_to_forbidden = getLatencyEstimate(getDistanceToPoly(neighbor_lon,neighbor_lat, forbiddenRegion))


                    print >> sys.stderr, '[%s;%s;%d] curr_id - %s, curr_lon - %s, curr_lat - %s, dist_to_forbidden - %f, dist_to_target - %f, neighbor_id - %s, neighbor_dist_to_target - %f, neighbor_dist_to_forbidden - %f, metric - %f'%(src_id, dst_id, STRATEGY, curr_node_id, curr_node_lon, curr_node_lat, dist_to_forbidden, dist_to_target, neighbor_id, l_dist_to_target, l_dist_to_target-l_min_distance, l_min_distance)

                        
                    for neighbor_id2 in nodes[neighbor_id].neighbors.keys():
                        if neighbor_id2 in processed:
                            continue

                        if neighbor_id2 not in nodes.keys():
                            continue

                        if nodes[neighbor_id].neighbors[neighbor_id2] < l_min_rtt_to_forbidden:
                            if nodes[neighbor_id2].coordinate.intersects(targetRegion):
                                print >> sys.stderr, '[%s;%s;%d] curr_id - %s, two-hop neighbor %s in target region, %s'%(src_id, dst_id, STRATEGY, curr_node_id, neighbor_id, str(path+[neighbor_id]))
                                if curr_node_id != src_id:
                                    qr_msg_len+=(20+len(neighbor_id2)+len(str(path)))
                                return True, path+[neighbor_id], 0, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len
                                found_someone_in_target = True
                                processed.append(neighbor_id)
                                break

                            neighbor_lon2,neighbor_lat2 = map((lambda array: array[0]), nodes[neighbor_id2].coordinate.coords.xy)

                            if dists_to_tr.has_key(neighbor_id2):
                                l2_dist_to_target = dists_to_tr[neighbor_id2]
                            else:
                                l2_dist_to_target = getDistanceToPoly(neighbor_lon2, neighbor_lat2, targetRegion)
                                dists_to_tr[neighbor_id2] = l2_dist_to_target

                            if dists_to_fr != None:
                                distance = l2_dist_to_target - dists_to_fr[nodes[neighbor_id2].coordinate_str]
                            else:
                                distance = l2_dist_to_target - getDistanceToPoly(neighbor_lon2,neighbor_lat2, forbiddenRegion)

                            if distance < l_min_distance:
                                l_min_distance = distance
                                l_min_node_id = neighbor_id2

                            print >> sys.stderr, '[%s;%s;%d] curr_id - %s, curr_lon - %s, curr_lat - %s, dist_to_forbidden - %f, dist_to_target - %f, two-hop neighbor_id - %s, two-hop neighbor_dist_to_target - %f, two-hop neighbor_dist_to_forbidden - %f, metric - %f'%(src_id, dst_id, STRATEGY, curr_node_id, curr_node_lon, curr_node_lat, dist_to_forbidden, dist_to_target, neighbor_id2, l2_dist_to_target, l2_dist_to_target-distance, distance)

                    if l_min_distance < min_distance:
                        min_distance = l_min_distance
                        min_node_id = neighbor_id

            if (min_node_id != None):
                if ttl != None and len(path) > ttl:
                    print >> sys.stderr, '[%s;%s;%d] curr_id - %s, ttl expired, %s'%(src_id, dst_id, STRATEGY, curr_node_id, str(path))
                    return False, path, 0, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len

                q_msg_len+=(q_msg_d_len+len(str(path)))
                path.append(min_node_id)
                to_process.append(min_node_id)
                q_msg_ctr+=1
            else:
                print >> sys.stderr, '[%s;%s;%d] curr_id - %s, no valid next hop, %s'%(src_id, dst_id, STRATEGY, curr_node_id, str(path))
                qr_msg_len+=(39+len(curr_node_id)+len(str(path)))
                return False, path, 0, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len

# Post-processing...
    if STRATEGY == 0:
        if ret_path == None:
            print >> sys.stderr, '[%s;%s;%d] curr_id - %s, routing failed'%(src_id, dst_id, STRATEGY, curr_node_id)
            return False, [], dc_msg_ctr, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len
        else:
            return True, ret_path, dc_msg_ctr, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len

    if STRATEGY == 1:
        print >> sys.stderr, '[%s;%s;%d] curr_id - %s, routing failed, %s'%(src_id, dst_id, STRATEGY, curr_node_id, str(path))

        return False, path, 0, q_msg_ctr, dists_to_tr, qr_msg_ctr, q_msg_len, qr_msg_len


