
 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

from shapely.wkt import dumps, loads
import string
import sys
from node import Node
from shapely.geometry import Point
import ast
import math
from random import sample, randint, seed

# Need files of the following format:
# COORDINATE_FILE:
# node_id longitude latitude

def loadNeighbors(neighbor_file):
    neighbors = dict()  # {src -> {nbr -> lat} }
    try:
        with open(neighbor_file) as f:
            for line in f:
                nodeId, neighborsString = line.split(";")
                try:
                    ls = ast.literal_eval(neighborsString)
                except:
                    #print neighborsString
                    #sys.exit(1)
                    continue
                d = dict()
                for (n,l) in ls:
                    d[n] = l
                neighbors[nodeId] = d
    except IOError:
        print "Could not read from file: %s" % neighbor_file
        sys.exit()

    return neighbors

def loadNodeInfo(coordinate_file, neighbor_file, is_random=False):
    nodes = dict()
    neighbors = loadNeighbors(neighbor_file)
    try:
        with open(coordinate_file) as f:
            for line in f:
                if line[0] == "#":
                    continue
                if is_random:
                    stuff = line.rstrip().split('\t')
                else:
                    stuff = line.split()
                node = Node()
                node.uniqueId = stuff[0]
                if stuff[0] == "None" or stuff[1] == "None":
                    continue
                node.coordinate = Point(float(stuff[1]), float(stuff[2]))
                node.coordinate_str = (stuff[1], stuff[2]) #ydlee
                if node.uniqueId in neighbors.keys() and len(neighbors[node.uniqueId]) > 0:
                    node.neighbors = neighbors[node.uniqueId]
                    nodes[node.uniqueId] = node
    except IOError:
        print "Could not read from file: %s" % coordinate_file
        sys.exit()
    return nodes

def compute_mean_and_std(data):
    mean = sum(data)*1. / len(data)

    std = math.sqrt(sum(map(lambda x: (x-mean)**2, data))/len(data))

    return mean, std


